require File.join(File.expand_path('../../../helpers', __FILE__),  'document_spec_helper')

class FooType < Struct.new(:bar)
  def self.to_tomte(value)
    'to_tomte'
  end

  def self.from_tomte(value)
    'from_tomte'
  end
end

describe 'A document Key' do
  Key = Tomte::Document::Keys::Key

  describe 'initializing a new key' do
    it 'should allow setting the name' do
      expect(Key.new(:foo, String).name).to eq 'foo'
    end

    it 'should allow setting the type' do
      expect(Key.new(:foo, Integer).type).to be(Integer)
    end

    it 'should allow setting options' do
      expect(Key.new(:foo, Integer, :required => true).options[:required]).to be(true)
    end

    it 'should default options to {}' do
      expect(Key.new(:foo, Integer, nil).options).to eq({})
    end

    it 'should symbolize option keys' do
      expect(Key.new(:foo, Integer, 'required' => true).options[:required]).to be(true)
    end

    it 'should work with just name' do
      key = Key.new(:foo)
      expect(key.name).to eq 'foo'
    end

    it 'should work with name and type' do
      key = Key.new(:foo, String)
      expect(key.name).to eq 'foo'
      expect(key.type).to eq String
    end

    it 'should work with name, type, and options' do
      key = Key.new(:foo, String, :required => true)
      expect(key.name).to eq 'foo'
      expect(key.type).to eq String
      expect(key.options[:required]).to be_truthy
    end

    it 'should work with name and options' do
      key = Key.new(:foo, :required => true)
      expect(key.name).to eq 'foo'
      expect(key.options[:required]).to be_truthy
    end
  end
  
  describe 'A key' do
    it 'should be equal to another key with same name and type' do
      expect(Key.new(:name, String)).to eq Key.new(:name, String)
    end

    it 'should not be equal to another key with different name' do
      expect(Key.new(:name, String)).not_to eq Key.new(:foo, String)
    end

    it 'should not be equal to another key with different type' do
      expect(Key.new(:name, String)).not_to eq Key.new(:name, Integer)
    end

    it 'should know if it is not a embedded_document' do
      expect(Key.new(:name, String).embeddable?).to be_falsey
    end

    it 'should know if it is a number' do
      expect(Key.new(:age, Integer).number?).to be_truthy
      expect(Key.new(:age, Float).number?).to be_truthy
    end

    it 'should know if it is not a number' do
      expect(Key.new(:age, String).number?).to be_falsey
    end
  end

  describe 'for an array with :typecast option' do
    before :each do
      @key = Key.new(:user_ids, Array, :typecast => 'Tomte::ObjectId')
    end
    subject { @key }

    it 'should cast each element correctly' do
      ids = [BSON::ObjectId.new, BSON::ObjectId.new, BSON::ObjectId.new.to_s, BSON::ObjectId.new.to_s]
      expect(subject.set(ids)).to eq ids.map { |id| Tomte::ObjectId.to_tomte(id) }
    end
  end

  describe 'for an array with :typecast option of Date' do
    before :each do
      @key = Key.new(:dates, Array, :typecast => 'Date')
    end 
    subject { @key }

    it 'should cast each element correctly when get' do
      dates = [Date.yesterday, Date.today, Date.tomorrow.to_s]
      expect(subject.get(dates)).to eq dates.map { |date| Date.from_tomte(date) }
    end

    it 'should cast each element correctly when set' do
      dates = [Date.yesterday, Date.today, Date.tomorrow.to_s]
      expect(subject.set(dates)).to eq dates.map { |date| Date.to_tomte(date) }
    end
  end

  describe 'for a set with :typecast option' do
    before :each do
      @key = Key.new(:user_ids, Set, :typecast => 'Tomte::ObjectId')
    end
    subject { @key }

    it 'should cast each element correctly' do
      ids = [BSON::ObjectId.new, BSON::ObjectId.new, BSON::ObjectId.new.to_s, BSON::ObjectId.new.to_s]
      expect(subject.set(ids)).to eq ids.map { |id| Tomte::ObjectId.to_tomte(id) }
    end
  end

  describe 'setting a value with a custom type' do
    it 'should correctly typecast' do
      key = Key.new(:foo, FooType)
      expect(key.set('something')).to eq 'to_tomte'
    end

    it 'should correctly typecast if object of that type is given' do
      key = Key.new(:foo, FooType)
      expect(key.set(FooType.new('something'))).to eq 'to_tomte'
    end
  end

  describe 'getting a value with a custom type' do
    it 'should use #from_tomte to convert back to custom type' do
      key = Key.new(:foo, FooType)
      expect(key.get('something')).to eq 'from_tomte'
    end
  end

  describe 'getting a value' do
    it 'should work with a type' do
      key = Key.new(:foo, String)
      expect(key.get('bar')).to eq 'bar'
    end

    it 'should work without type' do
      key = Key.new(:foo)
      expect(key.get([1, '2'])).to eq [1, '2']
      expect(key.get(false)).to eq false
      expect(key.get({})).to eq({})
    end
  end

  describe 'getting a value with a default set' do
    before :each do
      @key = Key.new(:foo, String, :default => 'baz')
    end

    it 'should return default value if value nil' do
      expect(@key.get(nil)).to eq 'baz'
    end

    it 'should return a dup of the default value' do
      @key.get(nil).replace('bar')
      expect(@key.get(nil)).to eq 'baz'
    end

    it 'should return value if not blank' do
      expect(@key.get('foobar')).to eq 'foobar'
    end

    it 'should work with Boolean type and false value' do
      expect(Key.new(:active, Boolean, :default => false).get(nil)).to be_falsey
    end

    it 'should work with Boolean type and true value' do
      expect(Key.new(:active, Boolean, :default => true).get(nil)).to be_truthy
    end

    it 'should work with procs' do
       expect(Key.new(:foo, String, :default => lambda { return 'hello world' }).get(nil)).to eq 'hello world'
    end
  end  
  
  
end


